function _imgObjectDraw(self)
  love.graphics.draw(self.img, self.x, self.y)
end

function _checkRectCollision(a, b)
  return a.x < b.x + b.w and
         a.x + a.w > b.x and
         a.y < b.y + b.h and
         a.h + a.y > b.y
end

-- get x,y,w,h of a rectangle that is AND of 2 rectangles
-- works only for colliding rects (check regular AABB collision first)
function _getOverlapingRect(a, b)
  -- offsets
  local x = math.max(a.x, b.x)
  local y = math.max(b.y, a.y)
  -- size
  local w = math.min(a.w + (a.x-b.x), a.w, (b.x+b.w) - a.x)
  local h = math.min(a.h + (a.y-b.y), a.h, (b.y+b.h) - a.y)

  return x, y, w, h
end

function _drawBBRect(a)
  love.graphics.setColor(255, 255, 255, 120)
  love.graphics.rectangle('line', a.x, a.y, a.w, a.h)
end
function _drawOverlappingRect(player, other)
  love.graphics.setColor(255, 0, 0)
  if _checkRectCollision(player, other) then
    love.graphics.rectangle('line', _getOverlapingRect(player, other))
  end
end

function _imgObjectCheckCollisionWith(self, other)
  -- check bounding box collision, if no bb collision then no pixel coolision either
  if not _checkRectCollision(self, other) then return false, false end

  local rx, ry, rw, rh = _getOverlapingRect(self, other)

  -- find overlaping non-transparent pixels
  local sx, sy, sdata = self.x, self.y, self.img_data
  local ox, oy, odata = other.x, other.y, other.img_data
  for x = rx, rx+rw-1 do
    for y = ry, ry+rh-1 do
      local dx = x - rx
      local dy = y - ry
      local _, _, _, sa = sdata:getPixel(rx-sx+dx, ry-sy+dy)
      local _, _, _, oa = odata:getPixel(rx-ox+dx, ry-oy+dy)
      if sa > 0 and oa > 0 then -- alpha > 0 is considered non-transparent, you might want a different value
        return true, true
      end
    end
  end
  return true, false
end

function newImgObject(name, x, y)
  local img_data = love.image.newImageData(name)
  local w, h = img_data:getDimensions()
  return {
    x = x, y = y,
    w = w, h = h,
    img = love.graphics.newImage(img_data),
    img_data = img_data,
    draw = _imgObjectDraw,
    checkCollisionWith = _imgObjectCheckCollisionWith
  }
end

function love.load()
  love.graphics.setBackgroundColor(0xFB, 0xCA, 0x80)
  love.window.setMode(480*2, 352*2)
  love.graphics.setDefaultFilter('nearest', 'nearest')
  map = newImgObject('map.png', 0, 0)
  player = newImgObject('player.png', 300, 130)
  alien = newImgObject('alien.png', 160, 154)
  tree = newImgObject('tree.png', 360, 141)
  is_drawing_boxes = true
end


function love.draw()
  love.graphics.setColor(255, 255, 255)
  love.graphics.scale(2, 2)
  map:draw()
  alien:draw()
  tree:draw()
  player:draw()

  if is_drawing_boxes then
    _drawBBRect(player)
    _drawBBRect(tree)
    _drawBBRect(alien)
    _drawOverlappingRect(player, tree)
    _drawOverlappingRect(player, alien)
  end

  love.graphics.setColor(0, 0, 0)
  love.graphics.origin()

  if player.collides_bb_tree then
    local txt = player.collides_pixel_tree and 'TREE PIXEL COLLISION' or 'TREE BB COLLISION'
    love.graphics.print(txt, 650, 500, 0, 2, 2)
  end

  if player.collides_bb_map then
    local txt = player.collides_pixel_map and 'MAP PIXEL COLLISION' or 'MAP BB COLLISION'
    love.graphics.print(txt, 300, 500, 0, 2, 2)
  end

  if player.collides_bb_alien then
    local txt = player.collides_pixel_alien and 'ALIEN PIXEL COLLISION' or 'ALIEN BB COLLISION'
    love.graphics.print(txt, 50, 320, 0, 2, 2)
  end

  local txt = 'FPS: ' .. love.timer.getFPS() .. '\n[space] - toggle boxes'
  love.graphics.printf(txt, 0, 10, love.graphics.getWidth()-10, 'right')
end

function love.update()
  player.collides_bb_tree, player.collides_pixel_tree = player:checkCollisionWith(tree)
  player.collides_bb_alien, player.collides_pixel_alien = player:checkCollisionWith(alien)
  player.collides_bb_map, player.collides_pixel_map = player:checkCollisionWith(map)

  if love.keyboard.isDown('left') then player.x = player.x - 1 end
  if love.keyboard.isDown('right') then player.x = player.x + 1 end
  if love.keyboard.isDown('up') then player.y = player.y - 1 end
  if love.keyboard.isDown('down') then player.y = player.y + 1 end
end

function love.keypressed(k)
  if k == 'space' then is_drawing_boxes = not is_drawing_boxes end
end
